#ifndef __BASICPLUGIN_H
#define __BASICPLUGIN_H

#include "volumeshop.h"

//#include <windows.h>

#include <dlfcn.h>
#define GetProcAddress(x, y) (dlsym(x,y))
#define FreeLibrary(x) (dlclose(x))
#define LoadLibrary(x) (dlopen(x,RTLD_NOW))
#define HINSTANCE void *
#define __declspec(x) 
#define GetTypeName(x) #x

#include <string>
#include <typeinfo>

#include "Plugin.h"
#include "Environment.h"
#include "Image.h"

#include <boost/filesystem/operations.hpp>
namespace fs = boost::filesystem;


#define DECLARE_PLUGIN(type,name) \
	extern "C" __declspec(dllexport) type * CreateInstance(Environment & envEnvironment); \
	extern "C" __declspec(dllexport) const char * GetType(); \
	extern "C" __declspec(dllexport) const char * GetName(); \
	extern "C" __declspec(dllexport) const char * GetDescription(); \
	extern "C" __declspec(dllexport) const unsigned char * GetIcon(); \

#define DEFINE_PLUGIN(type,pluginname,description,icon) \
	extern "C" __declspec(dllexport) type * CreateInstance(Environment & envEnvironment) \
	{ \
		return new pluginname(envEnvironment); \
	} \
	\
	extern "C" __declspec(dllexport) const char * GetType(void) \
	{ \
		return typeid(type).name(); \
	} \
	extern "C" __declspec(dllexport) const char * GetName(void) \
	{ \
		return #pluginname; \
	} \
	\
	extern "C" __declspec(dllexport) const char * GetDescription(void) \
	{ \
	return description; \
	} \
	\
	extern "C" __declspec(dllexport) const unsigned char * GetIcon(void) \
	{ \
	return icon; \
	} 
/*	\
	bool APIENTRY DllMain( HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved) \
	{ \
		return TRUE; \
	} \
*/

inline std::string GetPluginFilename(std::string pluginName)
{
	return "lib" + pluginName + ".so";
}


template <class Type>
class BasicPlugin : public Plugin
{
public:
	class Iterator
	{
	private:
		/* 
		 * Finds the next file with the given extension
		 * returns the name of the file found, or an empty string if at the end of the dir
		 */
		std::string FindNextFile()
		{
			fs::directory_iterator dirEnd;
			bool foundFile = false;
			std::string fileName("");

			while (m_itDir != dirEnd && !foundFile)
			{
				std::string curFile(m_itDir->leaf());
				if ( curFile.size() > m_strExtension.size() && curFile.substr(curFile.size() - m_strExtension.size()) == m_strExtension )
				{
					fileName = curFile;
					foundFile = true;

				}
				m_itDir++;

				bool isDone = dirEnd == m_itDir;
			}

			return fileName;

		}

		/*
		 * Finds the next file corresponding to the given module type
		 */
		std::string FindNextPluginFile()
		{
			std::string fileName("");
			bool foundFile = false;
			bool atEnd = false;

			while (!foundFile && ! atEnd)
			{
				std::string curFile(FindNextFile());
				if (curFile.size() == 0)
				{
					atEnd = true;
				}
				else
				{
					std::string strSearchType(typeid(Type).name());
					std::string strFoundType(GetType(curFile));
		
					

					if (strSearchType == strFoundType)
					{
						fileName = curFile;
						foundFile = true;
					}
				}
			}
		
			return fileName;
		}

	public:
		Iterator() : m_bAtEnd(false)
		{
			m_itDir = fs::directory_iterator( fs::current_path() );
			m_strExtension = "so";

			++(*this);
		};

		~Iterator()
		{
		};

		const Iterator & operator++()
		{
			m_strFilename = FindNextPluginFile();
			if (m_strFilename.size() == 0)
			{
				m_bAtEnd = true;
			}

			return *this;
		};

		const std::string operator*() const
		{
			return m_strFilename;
		};

		const std::string GetName() const
		{
			return BasicPlugin<Type>::GetName(m_strFilename);
		};

		const std::string GetDescription() const
		{
			return BasicPlugin<Type>::GetDescription(m_strFilename);
		};

		const Image GetIcon() const
		{
			return BasicPlugin<Type>::GetIcon(m_strFilename);
		};

		const bool IsAtEnd() const
		{
			return m_bAtEnd;
		};

	private:
		fs::directory_iterator m_itDir;
		std::string m_strExtension;
		std::string m_strFilename;
		
		bool m_bAtEnd;
	};
	
	BasicPlugin(Environment & envEnvironment, const std::string & strFilename) : Plugin(envEnvironment), m_strFilename(strFilename), m_hLibrary(NULL), m_pInstance(NULL)
	{
	};

	virtual ~BasicPlugin()
	{
		uninitialize();
	};

	virtual void initialize()
	{
		if (!IsInitialized())
		{
			typedef Type* (*CreateInstanceType)(Environment &);

			//Load the dll and keep the handle to it
			m_hLibrary = LoadLibrary(m_strFilename.c_str());

			// If the handle is valid, try to get the function address. 
			if (m_hLibrary != NULL) 
			{ 
				// Get pointer to our function using GetProcAddress
				CreateInstanceType CreateInstancePtr = (CreateInstanceType) GetProcAddress(m_hLibrary,"CreateInstance");

				// If the function address is valid, call the function. 
				if (CreateInstancePtr != NULL)
				{
					try
					{
						m_pInstance = CreateInstancePtr(GetEnvironment());
					}
					catch (Exception e)
					{
						if (m_pInstance)
						{
							try
							{
								delete m_pInstance;
							}
							catch (...)
							{
							}
						}

						m_pInstance = NULL;
						std::cerr << "Unable to initialize " << m_strFilename.c_str() << " (" << e << ")." << std::endl << std::endl;
					}
					catch (...)
					{
						if (m_pInstance)
						{
							try
							{
								delete m_pInstance;
							}
							catch (...)
							{
							}
						}

						m_pInstance = NULL;
						std::cerr << "Unable to initialize " << m_strFilename.c_str() << "." << std::endl << std::endl;
					}

					if (!m_pInstance)
					{
						FreeLibrary(m_hLibrary);
						m_pInstance = NULL;
					}
				}
				else
				{
					FreeLibrary(m_hLibrary);
					m_hLibrary = NULL;
				}
			}
		}
	};

	virtual void uninitialize()
	{
		if (IsInitialized())
		{
			if (m_hLibrary != NULL)
			{
				delete m_pInstance;
				m_pInstance = NULL;

				FreeLibrary(m_hLibrary);
				m_hLibrary = NULL;
			}
		}
	};

	virtual const bool IsInitialized() const
	{
		return (m_hLibrary != NULL && m_pInstance != NULL);
	};

	virtual const std::string GetType() const
	{
		if (IsInitialized())
		{
			std::string strResult;
			typedef const char* (*GetTypeType)();

			// If the handle is valid, try to get the function address. 
			if (m_hLibrary != NULL) 
			{ 
				// Get pointer to our function using GetProcAddress
				GetTypeType GetTypePtr = (GetTypeType) GetProcAddress(m_hLibrary,"GetType");

				// If the function address is valid, call the function. 
				if (GetTypePtr != NULL)
				{
					strResult = GetTypePtr();
				}
			}

			return strResult;
		}

		return GetType(m_strFilename);
	};

	virtual const std::string GetName() const
	{
		if (IsInitialized())
		{
			std::string strResult;
			typedef const char* (*GetNameType)();

			// If the handle is valid, try to get the function address. 
			if (m_hLibrary != NULL) 
			{ 
				// Get pointer to our function using GetProcAddress
				GetNameType GetNamePtr = (GetNameType) GetProcAddress(m_hLibrary,"GetName");

				// If the function address is valid, call the function. 
				if (GetNamePtr != NULL)
				{
					strResult = GetNamePtr();
				}
			}

			return strResult;
		}

		return GetName(m_strFilename);
	};

	virtual const std::string GetDescription() const
	{
		if (IsInitialized())
		{
			std::string strResult;
			typedef const char* (*GetDescriptionType)();

			// If the handle is valid, try to get the function address. 
			if (m_hLibrary != NULL) 
			{ 
				// Get pointer to our function using GetProcAddress
				GetDescriptionType GetDescriptionPtr = (GetDescriptionType) GetProcAddress(m_hLibrary,"GetDescription");

				// If the function address is valid, call the function. 
				if (GetDescriptionPtr != NULL)
				{
					strResult = GetDescriptionPtr();
				}
			}

			return strResult;
		}

		return GetDescription(m_strFilename);
	};

	virtual const Image GetIcon() const
	{
		if (IsInitialized())
		{
			Image imgResult;
			typedef const unsigned char* (*GetIconType)();

			// If the handle is valid, try to get the function address. 
			if (m_hLibrary != NULL) 
			{ 
				// Get pointer to our function using GetProcAddress
				GetIconType GetIconPtr = (GetIconType) GetProcAddress(m_hLibrary,"GetIcon");

				// If the function address is valid, call the function. 
				if (GetIconPtr != NULL)
				{
					const unsigned char *pIcon = GetIconPtr();

					if (pIcon)
					{
						const unsigned int uWidth(pIcon[0]);
						const unsigned int uHeight(pIcon[1]);
						const Color *pData = (Color *) (pIcon+2);
						imgResult = Image(uWidth,uHeight,pData);
					}
				}
			}

			return imgResult;
		}

		return GetIcon(m_strFilename);
	};

	static const std::string GetType(const std::string & strFilename)
	{
		std::string strResult;
		typedef const char* (*GetTypeType)();

		//Load the dll and keep the handle to it
		HINSTANCE hLibrary = LoadLibrary(strFilename.c_str());

		// If the handle is valid, try to get the function address. 
		if (hLibrary != NULL) 
		{ 
			// Get pointer to our function using GetProcAddress
			GetTypeType GetTypePtr = (GetTypeType) GetProcAddress(hLibrary,"GetType");

			// If the function address is valid, call the function. 
			if (GetTypePtr != NULL)
			{
				strResult = GetTypePtr();
			}

			FreeLibrary(hLibrary);
		}
		else
		{
			std::cout << "Could not load library file " << strFilename << std::endl;
			std::cout << "Error: " << dlerror() << std::endl;
		}

		return strResult;
	};

	static const std::string GetName(const std::string & strFilename)
	{
		std::string strResult;
		typedef const char* (*GetNameType)();

		//Load the dll and keep the handle to it
		HINSTANCE hLibrary = LoadLibrary(strFilename.c_str());

		// If the handle is valid, try to get the function address. 
		if (hLibrary != NULL) 
		{ 
			// Get pointer to our function using GetProcAddress
			GetNameType GetNamePtr = (GetNameType) GetProcAddress(hLibrary,"GetName");

			// If the function address is valid, call the function. 
			if (GetNamePtr != NULL)
			{
				strResult = GetNamePtr();
			}

			FreeLibrary(hLibrary);
		}

		return strResult;
	};

	static const std::string GetDescription(const std::string & strFilename)
	{
		std::string strResult;
		typedef const char* (*GetDescriptionType)();

		//Load the dll and keep the handle to it
		HINSTANCE hLibrary = LoadLibrary(strFilename.c_str());

		// If the handle is valid, try to get the function address. 
		if (hLibrary != NULL) 
		{ 
			// Get pointer to our function using GetProcAddress
			GetDescriptionType GetDescriptionPtr = (GetDescriptionType) GetProcAddress(hLibrary,"GetDescription");

			// If the function address is valid, call the function. 
			if (GetDescriptionPtr != NULL)
			{
				strResult = GetDescriptionPtr();
			}

			FreeLibrary(hLibrary);
		}

		return strResult;
	};

	static const Image GetIcon(const std::string & strFilename)
	{
		Image imgResult;
		typedef const unsigned char* (*GetIconType)();

		//Load the dll and keep the handle to it
		HINSTANCE hLibrary = LoadLibrary(strFilename.c_str());

		// If the handle is valid, try to get the function address. 
		if (hLibrary != NULL) 
		{ 
			// Get pointer to our function using GetProcAddress
			GetIconType GetIconPtr = (GetIconType) GetProcAddress(hLibrary,"GetIcon");

			// If the function address is valid, call the function. 
			if (GetIconPtr != NULL)
			{
				const unsigned char *pIcon = GetIconPtr();

				if (pIcon)
				{
					const unsigned int uWidth(pIcon[0]);
					const unsigned int uHeight(pIcon[1]);
					const Color *pData = (Color *) (pIcon+2);
					imgResult = Image(uWidth,uHeight,pData);
				}
			}

			FreeLibrary(hLibrary);
		}

		return imgResult;
	};

protected:

	Type * GetInstance()
	{
		return m_pInstance;
	};

private:

	const std::string m_strFilename;
	HINSTANCE m_hLibrary;
	Type *m_pInstance;
};

#endif

